import six
from ..qt_compat import QtGui, QtCore, QtWidgets

from .singleton import Singleton
from .model import DataModel

import logging as log

from six.moves.queue import Queue

import sys
import traceback


@Singleton
class Launcher(QtCore.QObject):

    pre = QtCore.pyqtSignal(str)
    post = QtCore.pyqtSignal()
    error = QtCore.pyqtSignal(str)
    result = QtCore.pyqtSignal(object)

    def __init__(self):
        QtCore.QObject.__init__(self)
        self.result.connect(self.on_result)
        self.current = None
        self.actionQueue = Queue(maxsize=1)
        self.actionQueue.put(True)
        self.forced = False
        self.cb = None

    def run(self, fn, cb=None, caption='Loading...', *args, **kwargs):
        self.actionQueue.get()

        self.cb = cb
        self.setup(caption)

        self.current = Action(fn, cb, *args, **kwargs)
        self.current.dummy.done.connect(self.on_result)
        self.current.dummy.error.connect(self.on_error)
        self.current.start()

    def setup(self, caption):
        log.info('\n### {} ###'.format(caption))
        self.pre.emit(caption)
        QtWidgets.qApp.processEvents()

    def cleanup(self):
        self.post.emit()
        QtWidgets.qApp.processEvents()

    def on_result(self, res=None):
        if self.cb is not None:
            self.cb(res)
        self.clear()

    def show_error(self, err):
        if type(err) == str:
            title = '[Error]'
        else:
            title = '[{}]'.format(type(err).__name__)

        try:
            traceback.print_last()
        except Exception as e:
            pass

        self.error.emit(title + '\n' + str(err))
        log.error(title + '\n' + str(err))

    def on_error(self, err):
        self.show_error(err)
        self.clear()

    def clear(self):
        self.current.dummy.done.disconnect()
        self.current.dummy.error.disconnect()
        del self.current
        self.actionQueue.put(True)
        self.cleanup()

    def terminate(self):
        pass

class Dummy(QtCore.QObject):
    done = QtCore.pyqtSignal(object)
    error = QtCore.pyqtSignal(object)

class Action(QtCore.QThread):

    dummy = Dummy()

    def __init__(self, fn, cb=None, *args, **kwargs):
        QtCore.QThread.__init__(self)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def run(self):
        log.info('\n### {} ###'.format(self.fn.__name__))
        try:
            res = self.fn(*self.args, **self.kwargs)
            self.dummy.done.emit(res)
        except Exception as e:
            self.dummy.error.emit(e)





import numpy as np

# 2D Dosition
#from scipy.linalg import svd
# 3D Dosition
#from sktensor import dtensor, cp_als

from ._convolutions import gconv, gconvs, gconvssh, ngconvssh


def make_rotated_grid(shape, orient):
    r = [s//2 for s in shape]
    r = [(-r[i], r[i]+1) if 2 * r[i] + 1 == shape[i] else (-r[i], r[i]) \
         for i in range(len(r))]
    z, y, x = np.mgrid[r[0][0]:r[0][1], r[1][0]:r[1][1], r[2][0]:r[2][1]]

    orgpts = np.vstack([z.ravel(), y.ravel(), x.ravel()])
    a, b = orient
    rotmz = [
        [1, 0, 0],
        [0, np.cos(a), -np.sin(a)],
        [0, np.sin(a), np.cos(a)]
    ]
    rotmy = [
        [np.cos(b), 0, np.sin(b)],
        [0, 1, 0],
        [-np.sin(b), 0, np.cos(b)]
    ]
    rotpts = np.dot(rotmy, np.dot(rotmz, orgpts))
    return [rotpts[i, :].reshape(z.shape) for i in range(3)]


def make_gaussian_1d(sigma=1., size=None, order=0, trunc=3):
    if size is None:
        size = sigma * trunc * 2 + 1
    x = np.arange(-(size//2), (size//2)+1)
    if order > 2:
        raise ValueError("Only orders up to 2 are supported")
    # compute unnormalized Gaussian response
    response = np.exp(-x ** 2 / (2. * sigma ** 2))
    if order == 1:
        response = -response * x
    elif order == 2:
        response = response * (x ** 2 - sigma ** 2)
    # normalize
    response /= np.abs(response).sum()
    return response.astype(np.float32)


def make_3d_gaussian(shape, sigma, orders=(0, 1, 0), orient=(0,0)):
    if isinstance(sigma, tuple) or isinstance(sigma, list):
        sz, sy, sx = sigma
    else:
        sz, sy, sx = (sigma * (1. + 2. * o) for o in orders)
    rotz, roty, rotx = make_rotated_grid(shape, orient)

    g = np.zeros(shape, dtype=np.float32)
    g[:] = np.exp(-0.5 * (rotx ** 2 / sx ** 2 + roty ** 2 / sy ** 2 + rotz ** 2 / sz ** 2))
    g /= 2 * np.pi * sx * sy * sz

    for o, s, x in zip(orders, (sz, sy, sx), (rotz, roty, rotx)):
        if o == 1:
            g *= -x
        elif o == 2:
            g *= (x**2 - s**2)

    return g


def make_3d_gabor(shape, sigmas, frequency, offset=0, orient=(0,0), return_real=True):
    sz, sy, sx = sigmas
    rotz, roty, rotx = make_rotated_grid(shape, orient)

    g = np.zeros(shape, dtype=np.complex)
    g[:] = np.exp(-0.5 * (rotx ** 2 / sx ** 2 + roty ** 2 / sy ** 2 + rotz ** 2 / sz ** 2))
    g /= 2 * np.pi * sx * sy * sz
    g *= np.exp(1j * (2 * np.pi * frequency * rotx + offset))

    if return_real:
        return np.real(g)

    return g


def rec_2d(D, weights=None):
    if weights is None:
        weights = np.ones(len(D), D[0][0].dtype)
    R = np.zeros((D[0][0].shape[0], D[0][1].shape[0]))
    for i, d in enumerate(D):
        R += d[0][:, None] * d[1][None, :] * weights[i]
    return R


def rec_3d(D, weights=None):
    if weights is None:
        weights = np.ones(len(D), D[0][0].dtype)
    R = np.zeros((D[0][0].shape[0], D[0][1].shape[0], D[0][2].shape[0]))
    for i, d in enumerate(D):
        R += d[0][:, None, None] * d[1][None, :, None] * d[2][None, None, :] \
             * weights[i]
    return R


def rec_error(X, R, mean=True):
    if mean:
        return np.mean((X - R)**2)
    else:
        return np.sum((X - R)**2)

"""def separate_kernel(kernel, max_rank=1, return_error=False, return_weights=False):
    if kernel.ndim == 1:
        return kernel
    elif kernel.ndim == 2:
        U, s, V = svd(kernel)

        mrank = np.linalg.matrix_rank(kernel)
        rank = max_rank if max_rank is not None and max_rank <= mrank  else mrank
        weights = s[:rank]

        if return_weights:
            D = [(U[:, i], V[i, :]) for i in range(rank)]
            if return_error:
                return D, weights, rec_error(kernel, rec_2d(D, weights))
            else:
                return D, weights
        else:
            D = [(U[:, i] * weights[i], V[i, :]) for i in range(rank)]
            if return_error:
                return D, rec_error(kernel, rec_2d(D))
            else:
                return D

    elif kernel.ndim == 3:
        rank = min(max_rank, min(*kernel.shape))
        T = dtensor(kernel)
        P, fit, itr, exectimes = cp_als(T, rank, init='random')

        Uz, Uy, Ux = P.U
        weights = P.lmbda

        if return_weights:
            D = [(Uz[:, i], Uy[:, i], Ux[:, i]) for i in range(rank)]
            if return_error:
                return D, weights, rec_error(kernel, rec_3d(D, weights))
            else:
                return D, weights
        else:
            D = [(Uz[:, i] * weights[i], Uy[:, i], Ux[:, i]) for i in range(rank)]
            if return_error:
                return D, rec_error(kernel, rec_3d(D))
            else:
                return D
    else:
        raise Exception("Kernel dimensions not supported")"""





import numpy as np
from skimage import color
from scipy import ndimage

from _features import _position_3d, _position_2d


class DensePositionExtractor(object):
    """Extracts position from images"""

    def __init__(self, unit=True, multichannel=False, **kwargs):
        self.unit = unit
        self.multichannel = multichannel

    def transform(self, X):
        if self.multichannel:
            if X.ndim == 3:
                G = _position_2d(X)
                G.shape = (-1, 2)
            else:
                G = _position_3d(X)
                G.shape = (-1, 3)
        else:
            if X.ndim == 2:
                G = _position_2d(X)
                G.shape = (-1, 2)
            else:
                G = _position_3d(X)
                G.shape = (-1, 3)

        if self.unit:
            G /= G.max(axis=0)

        return G


class DenseColorExtractor(object):
    """Extracts intensity features from images (grayscale or color)"""

    def __init__(self, multichannel=False, convert2lab=False, **kwargs):
        self.multichannel = multichannel
        self.convert2lab = convert2lab

    def transform(self, X):
        if self.multichannel and self.convert2lab:
            X = color.rgb2lab(X)

        if self.multichannel:
            return X.reshape(-1, X.shape[-1])
        else:
            return X.reshape(-1, 1)


class DenseTextureExtractor(object):
    """Extracts local texture features from images"""

    def __init__(self, multichannel=False, order=2, sigmas=[1., 2.], **kwargs):
        self.multichannel = multichannel
        self.order = order
        self.sigmas = sigmas

    def transform(self, X):
        f = []
        for sigma in self.sigmas:
            g = ndimage.gaussian_filter(X, sigma)
            dg = np.gradient(g)
            g.shape = g.shape + (1,)
            for d in dg:
                d.shape = d.shape + (1,)
            f += [g,] + dg
        f = np.concatenate(f, axis=-1)
        f.shape = (-1, f.shape[-1])
        return f


class DensePatchExtractor2D(object):

    def __init__(self, patch_size=(21,21), flatten=True):
        self.patch_size = patch_size
        self.flatten = flatten

    def transform(self, img):
        ypad, xpad = self.patch_size[0]//2, self.patch_size[1]//2
        padded = np.pad(img, ((ypad, ypad), (xpad, xpad)), mode='reflect')
        Y, X = padded.shape
        y, x = self.patch_size
        shape = ((Y-y+1), (X-x+1), y, x) # number of patches, patch_shape
        # The right strides can be thought by:
        # 1) Thinking of `img` as a chunk of memory in C order
        # 2) Asking how many items through that chunk of memory are needed when indices
        #    i,j,k,l are incremented by one
        strides = padded.itemsize*np.array([X, 1, X, 1])
        if self.flatten:
            return np.lib.stride_tricks.as_strided(padded, shape=shape, strides=strides).reshape(-1, self.patch_size[0] * self.patch_size[1])
        else:
            return np.lib.stride_tricks.as_strided(padded, shape=shape, strides=strides)


class DensePatchExtractor3D(object):

    def __init__(self, patch_size=21, flatten=False, pad=True):
        self.patch_size = patch_size
        self.flatten = flatten
        self.pad = pad

    def transform(self, img):
        if self.pad:
            padded = np.pad(img, self.patch_size//2, mode='reflect')
        else:
            padded = img.copy()
        Z, Y, X = padded.shape
        z, y, x = self.patch_size, self.patch_size, self.patch_size
        shape = ((Z-z+1), (Y-y+1), (X-x+1), z, y, x) # number of patches, patch_shape
        strides = padded.itemsize*np.array([Y*X, X, 1, Y*X, X, 1])
        if self.flatten:
            return np.lib.stride_tricks.as_strided(padded, shape=shape, strides=strides).reshape(-1, self.patch_size[0] * self.patch_size[1])
        else:
            return np.lib.stride_tricks.as_strided(padded, shape=shape, strides=strides)
